@@ -1,6 +1,7 @@
 import json
-import urllib.request
+import shutil
 import time
+import urllib.request
 
 
 class BaseGTmetrixAPIException(Exception):
@@ -41,10 +42,11 @@ class Requestor:
         auth_handler = urllib.request.HTTPBasicAuthHandler(password_manager)
 
         self._opener = urllib.request.build_opener(auth_handler, NoRedirect)
+        self._redirect_opener = urllib.request.build_opener(auth_handler)
         self._sleep = sleep_function
 
-    def _plain_request(self, url, **kwargs):
-        # method=None, data=None, headers={}, valid_status=None, valid_statuses=None, require_data=True):
+    def _plain_request(self, url, opener, **kwargs):
+        # method=None, data=None, headers={}, valid_status=None, valid_statuses=None, return_data=True, require_data=True, return_json=True
         data = kwargs.get("data", None)
         if isinstance(data, dict):
             data = json.dumps(data)
@@ -57,7 +59,7 @@ class Requestor:
             method=kwargs.get("method", None),
         )
         try:
-            response = self._opener.open(request)
+            response = opener.open(request)
         except urllib.error.HTTPError as e:
             # https://docs.python.org/3/library/urllib.error.html#urllib.error.HTTPError
             # > Though being an exception (a subclass of URLError), an HTTPError
@@ -77,12 +79,17 @@ class Requestor:
                     response,
                     response.read(),
                 )
+        if not kwargs.get("return_data", True):
+            return (response, None)
         data = response.read()
+        # TODO: retry reads if you can't read the whole response in one go
         if len(data) == 0:
             if kwargs.get("require_data", True):
                 raise GTmetrixAPIFailureException("API returned empty response", request, response, data)
             else:
                 return (response, None)
+        # if not kwargs.get("return_json", True):
+        #     return (response, data)
         try:
             json_data = json.loads(data.decode())
         except (json.JSONDecodeError, UnicodeError) as e:
@@ -134,8 +141,12 @@ class Requestor:
                 self._sleep(delay)
                 return self._retry_request(url, retries - 1, **kwargs)
 
-    def request(self, url, data=None, **kwargs):
-        return self._retry_request(url, data=data, **kwargs)
+    def request(self, url, follow_redirects=False, data=None, **kwargs):
+        if follow_redirects:
+            opener = self._redirect_opener
+        else:
+            opener = self._opener
+        return self._retry_request(url, opener=opener, data=data, **kwargs)
 
 
 def dict_is_error(data):
@@ -173,6 +184,19 @@ def dict_is_report(data):
     )
 
 
+def dict_is_user(data):
+    return (
+        isinstance(data, dict)
+        and "type" in data
+        and data["type"] == "user"
+        and "id" in data
+        and "attributes" in data
+        and isinstance(data["attributes"], dict)
+        and "api_credits" in data["attributes"]
+        and "api_refill" in data["attributes"]
+    )
+
+
 class Object(dict):
     def __init__(self, requestor, data, sleep_function=time.sleep):
         super().__init__(**data)
@@ -251,6 +275,49 @@ class Report(Object):
                 )
         return Report(requestor, response_data["data"], sleep_function)
 
+    def delete(self):
+        self._requestor.request("reports/" + self["id"], method="DELETE", return_data=False)
+
+    def retest(self):
+        (response, response_data) = self._requestor.request("reports/%s/retest" % self["id"], method="POST")
+        # TODO: this is same as when starting a test
+        if __debug__:
+            if not "data" in response_data:
+                raise GTmetrixAPIFailureException(
+                    "API returned no data for a retest",
+                    None,
+                    response,
+                    response_data,
+                )
+            if not dict_is_test(response_data["data"]):
+                raise GTmetrixAPIFailureException(
+                    "API returned non-test for a retest",
+                    None,
+                    response,
+                    response_data,
+                )
+        test = Test(self._requestor, response_data["data"])
+        return test
+
+    def getresource(self, name, destination=None):
+        (response, response_data) = self._requestor.request(
+            "reports/%s/resources/%s" % (self["id"], name),
+            follow_redirects=True,
+            return_data=False,
+        )
+        if destination is None:
+            data = b""
+            chunk = response.read()
+            while chunk:
+                data += chunk
+                chunk = response.read()
+            return data
+        elif isinstance(destination, str):
+            with open(destination, "wb") as destination_file:
+                shutil.copyfileobj(response, destination_file)
+        else:
+            shutil.copyfileobj(response, destination)
+
 
 class Interface:
     def __init__(
@@ -259,15 +326,15 @@ class Interface:
         base_url="https://gtmetrix.com/api/2.0/",
         sleep_function=time.sleep,
     ):
-        self.requestor = Requestor(api_key, base_url, sleep_function)
+        self._requestor = Requestor(api_key, base_url, sleep_function)
 
     def start_test(self, url, **attributes):
         """Start a Test"""
         attributes["url"] = url
         data = {"type": "test", "attributes": attributes}
-        (response, response_data) = self.requestor.request(
+        (response, response_data) = self._requestor.request(
             "tests",
-            {"data": data},
+            data={"data": data},
             method="POST",
             headers={"Content-Type": "application/vnd.api+json"},
         )
@@ -286,7 +353,7 @@ class Interface:
                     response,
                     response_data,
                 )
-        test = Test(self.requestor, response_data["data"])
+        test = Test(self._requestor, response_data["data"])
         # TODO: do something with credits_left and credits_used
         return test
 
@@ -305,22 +372,14 @@ class Interface:
         if filter is not None:
             query.extend(["filter[%s]=%s" % (k, v) for (k, v) in filter.items()])
 
-        (response, response_data) = self.requestor.request("tests?" + "&".join(query))
+        (response, response_data) = self._requestor.request("tests?" + "&".join(query))
         # TODO: pagination:
-        # if there is link[next]:
-        #   repeat with page_number=N+1
-        # pagination v2:
-        # if there is link[next], then:
-        #  repeat with page[number]=2 (up to 100th)
-        #  if there is link[next], then
-        #    repeat with page[size=100]&page[number]=2 (up to 200th)
-        #    if there is link[next], then
-        #      repeat with page[size=200]&page[number]=2 (up to 400th)
-        #      if there is link[next], then
-        #        repeat with page[size=400]&page[number]=2 (up to 800th)
-        #        # recursion stops here
-        #        while there is link[next]:
-        #          repeat with page[size=400]&page[number]=(N++)
+        # next_link=first_link
+        # results=[]
+        # while next_link:
+        #  request
+        #  results.extend(request_data...)
+        #  next_link=request_data.get(..., None)
         if __debug__:
             if not "data" in response_data:
                 raise GTmetrixAPIFailureException(
@@ -343,5 +402,19 @@ class Interface:
                     response,
                     response_data,
                 )
-        tests = [Test(self.requestor, test_data) for test_data in response_data["data"]]
+        tests = [Test(self._requestor, test_data) for test_data in response_data["data"]]
         return tests
+
+    def status(self):
+        (response, response_data) = self._requestor.request("status")
+        if __debug__:
+            if not "data" in response_data:
+                raise GTmetrixAPIFailureException("API returned no data for status", None, response, response_data)
+            if not dict_is_user(response_data["data"]):
+                raise GTmetrixAPIFailureException(
+                    "API returned non-user for status",
+                    None,
+                    response,
+                    response_data,
+                )
+        return response_data["data"]
